"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const winston = require("winston");
const uml = require("./uml/index");
class Delinter {
    /**
     * Uml code model description filled by the parse method(s)
     *
     * @readonly
     * @type {uml.CodeModel}
     * @memberOf Delinter
     */
    get umlCodeModel() {
        return this._umlCodeModel;
    }
    constructor() {
        this._umlCodeModel = new uml.CodeModel();
    }
    /**
     * Delint a TypeScript source file, adding the parsed elements to umlCodeModel.
     *
     * @param {ts.SourceFile} file TypeScript source file
     *
     * @memberOf Delinter
     */
    parse(file) {
        this._delintNode(file);
    }
    _delintNode(node) {
        winston.log("debug", "delintNode", { "node.kind": node.kind, "ts.SyntaxKind": ts.SyntaxKind[node.kind] });
        switch (node.kind) {
            case ts.SyntaxKind.ClassDeclaration:
                this._delintClass(node);
                break;
            case ts.SyntaxKind.InterfaceDeclaration:
                this._delintInterface(node);
                break;
            default:
                ts.forEachChild(node, (n) => { this._delintNode(n); });
                break;
        }
    }
    _delintClassNode(node, umlClass) {
        winston.log("debug", "delintClassNode", { "node.kind": node.kind, "ts.SyntaxKind": ts.SyntaxKind[node.kind] });
        switch (node.kind) {
            case ts.SyntaxKind.PropertyDeclaration:
                this._delintClassProperty(node, umlClass);
                break;
            case ts.SyntaxKind.GetAccessor:
            case ts.SyntaxKind.SetAccessor:
                this._delintClassGetterSetter(node, umlClass);
                break;
            case ts.SyntaxKind.MethodDeclaration:
                this._delintClassMethod(node, umlClass);
                break;
            default:
                ts.forEachChild(node, (n) => { this._delintClassNode(n, umlClass); });
                break;
        }
    }
    _delintClass(node) {
        const umlClass = new uml.Class(node.name.getText());
        this._umlCodeModel.nodes.setValue(umlClass.identifier, umlClass);
        this._delintHeritageClauses(node.heritageClauses, umlClass);
        ts.forEachChild(node, (n) => { this._delintClassNode(n, umlClass); });
    }
    _delintInterface(node) {
        const umlInterface = new uml.Class(node.name.getText(), 2 /* Interface */);
        this._umlCodeModel.nodes.setValue(umlInterface.identifier, umlInterface);
        this._delintHeritageClauses(node.heritageClauses, umlInterface);
        ts.forEachChild(node, (n) => { this._delintClassNode(n, umlInterface); });
    }
    _delintHeritageClauses(heritageClauses, umlClass) {
        if (heritageClauses) {
            heritageClauses.forEach((h) => {
                winston.log("debug", "delintHeritageClauses", { "h.token": h.token, "ts.SyntaxKind": ts.SyntaxKind[h.token] });
                switch (h.token) {
                    case ts.SyntaxKind.ImplementsKeyword:
                        h.types.forEach((t) => {
                            const interfaceName = t.expression.getText();
                            // Add interface to CodeModel if not exists
                            if (!this._umlCodeModel.nodes.containsKey(interfaceName)) {
                                const umlInterface = new uml.Class(interfaceName, 2 /* Interface */);
                                this._umlCodeModel.nodes.setValue(interfaceName, umlInterface);
                            }
                            const generalization = new uml.Generalization(umlClass.identifier, interfaceName);
                            this._umlCodeModel.generalizations.add(generalization);
                        });
                        break;
                    case ts.SyntaxKind.ExtendsKeyword:
                        h.types.forEach((t) => {
                            const parentClassName = t.expression.getText();
                            // Add interface to CodeModel if not exists
                            if (!this._umlCodeModel.nodes.containsKey(parentClassName)) {
                                const umlParentClass = new uml.Class(parentClassName);
                                this._umlCodeModel.nodes.setValue(parentClassName, umlParentClass);
                            }
                            const generalization = new uml.Generalization(umlClass.identifier, parentClassName);
                            this._umlCodeModel.generalizations.add(generalization);
                        });
                        break;
                    /* istanbul ignore next: default case never reached */
                    default:
                        break;
                }
            });
        }
    }
    _delintClassMethod(methodDeclaration, umlClass) {
        const identifier = methodDeclaration.name.getText();
        // Default to public accessibility
        const accessibility = this._delintAccessibilityModifiers(methodDeclaration.modifiers);
        const method = new uml.FunctionProperty(identifier, accessibility);
        if (methodDeclaration.type) {
            method.returnType = this._delintType(methodDeclaration.type);
        }
        methodDeclaration.parameters.forEach((p) => {
            const parameter = new uml.Parameter(p.name.getText());
            parameter.type = this._delintType(p.type);
            if (p.initializer) {
                parameter.defaultInitializer = p.initializer.getText();
                parameter.optional = true;
            }
            if (p.questionToken) {
                parameter.optional = true;
            }
            method.parameters.push(parameter);
        });
        umlClass.methods.setValue(identifier, method);
    }
    _delintClassProperty(property, umlClass) {
        const identifier = property.name.getText();
        // Default to public accessibility
        const accessibility = this._delintAccessibilityModifiers(property.modifiers);
        const type = this._delintType(property.type);
        const variable = new uml.VariableProperty(identifier, accessibility, type);
        umlClass.variables.setValue(identifier, variable);
        this._typeAssociations(umlClass.identifier, type).forEach((a) => {
            this.umlCodeModel.associations.add(a);
        });
    }
    _delintClassGetterSetter(node, umlClass) {
        const identifier = node.name.getText();
        // Default to public accessibility
        const accessibility = this._delintAccessibilityModifiers(node.modifiers);
        let type;
        if (node.kind === ts.SyntaxKind.GetAccessor) {
            type = this._delintType(node.type);
        }
        else {
            if (node.parameters.length > 0) {
                type = this._delintType(node.parameters[0].type);
            }
        }
        const variable = new uml.VariableProperty(identifier, accessibility, type);
        // Set stereotype based on kind
        const existingVariable = umlClass.variables.getValue(identifier);
        let stereotype;
        if (node.kind === ts.SyntaxKind.GetAccessor) {
            // If a setter already exists, use stereotype GetSet
            if (existingVariable && existingVariable.stereotype === 4 /* Set */) {
                stereotype = 5 /* GetSet */;
            }
            else {
                stereotype = 3 /* Get */;
            }
        }
        else {
            // If a setter already exists, use stereotype GetSet
            if (existingVariable && existingVariable.stereotype === 3 /* Get */) {
                stereotype = 5 /* GetSet */;
            }
            else {
                stereotype = 4 /* Set */;
            }
        }
        variable.stereotype = stereotype;
        umlClass.variables.setValue(identifier, variable);
        this._typeAssociations(umlClass.identifier, type).forEach((a) => {
            this.umlCodeModel.associations.add(a);
        });
    }
    _delintAccessibilityModifiers(modifiers) {
        let accessibility = 0 /* Public */;
        if (modifiers) {
            modifiers.forEach((m) => {
                winston.log("debug", "delintClassProperty", { "m.kind": m.kind, "ts.SyntaxKind": ts.SyntaxKind[m.kind] });
                switch (m.kind) {
                    case ts.SyntaxKind.PrivateKeyword:
                        accessibility = 2 /* Private */;
                        break;
                    case ts.SyntaxKind.ProtectedKeyword:
                        accessibility = 1 /* Protected */;
                        break;
                    case ts.SyntaxKind.PublicKeyword:
                        accessibility = 0 /* Public */;
                        break;
                }
            });
        }
        return accessibility;
    }
    _delintType(typeNode) {
        let type = null;
        const kind = typeNode ? typeNode.kind : null;
        switch (kind) {
            case ts.SyntaxKind.AnyKeyword:
            case ts.SyntaxKind.NumberKeyword:
            case ts.SyntaxKind.BooleanKeyword:
            case ts.SyntaxKind.StringKeyword:
            case ts.SyntaxKind.SymbolKeyword:
            case ts.SyntaxKind.VoidKeyword:
                type = new uml.PrimaryType(typeNode.getText(), 1 /* PredefinedType */);
                break;
            case ts.SyntaxKind.TypeReference:
                const typeReference = typeNode;
                type = new uml.PrimaryType(typeReference.getText(), 2 /* TypeReference */);
                type.name = typeReference.typeName.getText();
                const typeRef = typeNode;
                type.typeArguments = this._delintTypeArguments(typeRef.typeArguments);
                break;
            case ts.SyntaxKind.TypeLiteral:
                type = new uml.PrimaryType("TypeLiteral", 3 /* ObjectType */);
                break;
            case ts.SyntaxKind.TupleType:
                type = new uml.PrimaryType(typeNode.getText(), 5 /* TupleType */);
                const tupleType = typeNode;
                type.typeArguments = this._delintTypeArguments(tupleType.elementTypes);
                break;
            case ts.SyntaxKind.ArrayType:
                type = new uml.PrimaryType(typeNode.getText(), 4 /* ArrayType */);
                break;
            case ts.SyntaxKind.TypeQuery:
                type = new uml.PrimaryType(typeNode.getText(), 6 /* TypeQuery */);
                break;
            case ts.SyntaxKind.ThisType:
                type = new uml.PrimaryType(typeNode.getText(), 7 /* ThisType */);
                break;
            case ts.SyntaxKind.UnionType:
                type = new uml.UnionOrIntersectionType(typeNode.getText(), 100 /* Union */);
                const union = typeNode;
                type.types = this._delintTypeArguments(union.types);
                break;
            case ts.SyntaxKind.IntersectionType:
                type = new uml.UnionOrIntersectionType(typeNode.getText(), 101 /* Intersection */);
                const intersection = typeNode;
                type.types = this._delintTypeArguments(intersection.types);
                break;
            default:
                type = new uml.PrimaryType("any", 0 /* ImplicitAny */);
                break;
        }
        return type;
    }
    _delintTypeArguments(typeArguments) {
        let delintedTypes = [];
        if (typeArguments) {
            delintedTypes = typeArguments.map((value) => {
                return this._delintType(value);
            });
        }
        return delintedTypes;
    }
    _typeAssociations(from, type) {
        // Add association to type
        let associations = [];
        if (type) {
            if (type instanceof uml.PrimaryType || type instanceof uml.UnionOrIntersectionType) {
                switch (type.kind) {
                    case 2 /* TypeReference */:
                        associations = [new uml.Association(from, type.name)];
                    // Fall through to process potential (generic) type arguments
                    case 4 /* ArrayType */:
                    case 5 /* TupleType */:
                        type.typeArguments.forEach((t) => {
                            associations = associations.concat(this._typeAssociations(from, t));
                        });
                        break;
                    case 100 /* Union */:
                    case 101 /* Intersection */:
                        type.types.forEach((t) => {
                            associations = associations.concat(this._typeAssociations(from, t));
                        });
                        break;
                    default:
                        // Do not produce associations
                        break;
                }
            }
        }
        return associations;
    }
}
exports.Delinter = Delinter;
