"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const uml = require("../uml/index");
const formatter_1 = require("./formatter");
/**
 * PlantUML Formatter
 *
 * http://plantuml.com
 *
 * @export
 * @class Formatter
 * @extends {AbstractFormatter}
 */
class Formatter extends formatter_1.AbstractFormatter {
    generateClassDiagram(umlCodeModel) {
        let content;
        content = this._formatNodes(umlCodeModel);
        content += "\n";
        content += this._formatLinks(umlCodeModel);
        if (this.options.plantuml.diagramTags) {
            return `@startuml\n${content}\n@enduml\n`;
        }
        else {
            return content;
        }
    }
    _formatNodes(umlCodeModel) {
        const content = [];
        umlCodeModel.nodes.forEach((key, node) => {
            if (node instanceof uml.Class) {
                const properties = this._formatProperties(node);
                switch (node.stereotype) {
                    case 2 /* Interface */:
                        content.push(`interface ${node.identifier}{\n${properties}\n}\n`);
                        break;
                    case 0 /* None */:
                    default:
                        content.push(`class ${node.identifier}{\n${properties}\n}\n`);
                        break;
                }
            }
        });
        return content.join("\n");
    }
    _formatProperties(node) {
        const properties = [];
        node.variables.forEach((key, variable) => {
            properties.push(this._formatVariable(variable));
        });
        node.methods.forEach((key, method) => {
            properties.push(this._formatMethod(method));
        });
        return properties.join("\n");
    }
    _formatVariable(variable) {
        const accessibility = this._formatAccessibility(variable.accessibility);
        return `  ${accessibility}${variable.identifier} : ${variable.type.text}`;
    }
    _formatMethod(method) {
        const accessibility = this._formatAccessibility(method.accessibility);
        let returns = "";
        if (method.returnType) {
            returns = `: ${method.returnType.text}`;
        }
        const parameters = method.parameters.map((p) => {
            let type = "";
            let initializer = "";
            let questionMark = "";
            if (p.type) {
                type = `: ${p.type.text}`;
            }
            if (p.defaultInitializer) {
                initializer = ` = ${p.defaultInitializer}`;
            }
            else if (p.optional) {
                questionMark = "?";
            }
            return `${p.identifier}${questionMark}${type}${initializer}`;
        });
        return `  ${accessibility}${method.identifier}(${parameters.join(", ")})${returns}`;
    }
    _formatLinks(umlCodeModel) {
        const content = [];
        umlCodeModel.associations.forEach((link) => {
            const reverse = link.reverse();
            if (umlCodeModel.associations.contains(reverse)) {
                if (link.toString().localeCompare(reverse.toString()) < 0) {
                    content.push(`${link.fromName} -- ${link.toName}`);
                }
            }
            else {
                content.push(`${link.fromName} --> ${link.toName}`);
            }
        });
        umlCodeModel.generalizations.forEach((link) => {
            content.push(`${link.toName} <|-- ${link.fromName}`);
        });
        return content.join("\n");
    }
    _formatAccessibility(accessibility) {
        let result = "";
        switch (accessibility) {
            default:
            case 0 /* Public */:
                result = "+";
                break;
            case 1 /* Protected */:
                result = "#";
                break;
            case 2 /* Private */:
                result = "-";
                break;
        }
        return result;
    }
}
exports.Formatter = Formatter;
