"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typescript_collections_1 = require("typescript-collections");
const uml = require("../uml/index");
const formatter_1 = require("./formatter");
/**
 * yUML Formatter
 *
 * https://yuml.me/diagram/scruffy/class/samples
 *
 * @export
 * @class Formatter
 * @extends {AbstractFormatter}
 */
class Formatter extends formatter_1.AbstractFormatter {
    constructor(options) {
        super(options);
        this._outputtedNodes = new typescript_collections_1.Set();
    }
    generateClassDiagram(umlCodeModel) {
        let yuml = "// {type:class}\n";
        this._outputtedNodes.clear();
        umlCodeModel.generalizations.forEach((value) => {
            yuml += this._formatNode(umlCodeModel.nodes.getValue(value.toName));
            yuml += "^" + this._formatNode(umlCodeModel.nodes.getValue(value.fromName)) + "\n";
            this._outputtedNodes.add(value.fromName);
            this._outputtedNodes.add(value.toName);
        });
        umlCodeModel.associations.forEach((link) => {
            const reverse = link.reverse();
            let formattedAssociation = null;
            if (umlCodeModel.associations.contains(reverse)) {
                if (link.toString().localeCompare(reverse.toString()) < 0) {
                    formattedAssociation = this._formatNode(umlCodeModel.nodes.getValue(link.fromName));
                    formattedAssociation += "-" + this._formatNode(umlCodeModel.nodes.getValue(link.toName)) + "\n";
                }
            }
            else {
                formattedAssociation = this._formatNode(umlCodeModel.nodes.getValue(link.fromName));
                formattedAssociation += "->" + this._formatNode(umlCodeModel.nodes.getValue(link.toName)) + "\n";
            }
            if (formattedAssociation) {
                yuml += formattedAssociation;
                this._outputtedNodes.add(link.fromName);
                this._outputtedNodes.add(link.toName);
            }
        });
        umlCodeModel.nodes.forEach((key, value) => {
            if (!this._outputtedNodes.contains(key)) {
                yuml += this._formatNode(value) + "\n";
            }
        });
        return yuml;
    }
    _formatNode(node) {
        if (node instanceof uml.Class) {
            const properties = this._formatProperties(node);
            switch (node.stereotype) {
                case 2 /* Interface */:
                    return `[<<${node.identifier}>>|${properties}]`;
                default:
                    return `[${node.identifier}|${properties}]`;
            }
        }
    }
    _formatProperties(node) {
        const variables = node.variables.values().map((variable) => {
            return this._formatVariable(variable);
        }).join(";");
        const methods = node.methods.values().map((method) => {
            return this._formatMethod(method);
        }).join(";");
        return this._replaceSpecialCharacters(`${variables}|${methods}`);
    }
    _formatVariable(variable) {
        const accessibility = this._formatAccessibility(variable.accessibility);
        return `${accessibility}${variable.identifier}:${variable.type.text}`;
    }
    _formatMethod(method) {
        const accessibility = this._formatAccessibility(method.accessibility);
        let returns = "";
        if (method.returnType) {
            returns = `:${method.returnType.text}`;
        }
        const parameters = method.parameters.map((p) => {
            let type = "";
            let initializer = "";
            let questionMark = "";
            if (p.type) {
                type = `:${p.type.text}`;
            }
            if (p.defaultInitializer) {
                initializer = `=${p.defaultInitializer}`;
            }
            else if (p.optional) {
                questionMark = "?";
            }
            return `${p.identifier}${questionMark}${type}${initializer}`;
        });
        return `${accessibility}${method.identifier}(${parameters.join(",")})${returns}`;
    }
    _formatAccessibility(accessibility) {
        let result = "";
        switch (accessibility) {
            default:
            case 0 /* Public */:
                result = "+";
                break;
            case 1 /* Protected */:
                result = "#";
                break;
            case 2 /* Private */:
                result = "-";
                break;
        }
        return result;
    }
    _replaceSpecialCharacters(value) {
        return value
            .replace(/\[/g, "［")
            .replace(/\]/g, "］");
    }
}
exports.Formatter = Formatter;
