"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = require("fs");
const ts = require("typescript");
const delint_1 = require("./delint");
const formatter = require("./formatter/index");
const uml = require("./uml/index");
class TypeScriptUml {
    constructor() {
        // Static class, don't allow instantiation
    }
    /**
     * Parse the source files in a TypeScript project
     *
     * @static
     * @param {string} rootPath Project root path, if tsConfigPath is not defined, the tsconfig.json file
     * will be searched in this directory
     * @param {string} [tsConfigPath] (Optional) Path to tsconfig.json file
     * @returns {uml.CodeModel} The parse results
     *
     * @memberOf TypeScriptUml
     */
    static parseProject(rootPath, options) {
        const _options = this._setDefaultParseOptions(options);
        const tsConfig = this._readTsconfig(rootPath, _options);
        const delinter = new delint_1.Delinter();
        for (const f of tsConfig.fileNames) {
            this.parseFile(f, tsConfig.options.target, undefined, delinter);
        }
        return delinter.umlCodeModel;
    }
    /**
     * Parse a single TypeScript source file
     *
     * @static
     * @param {string} fileName Source file to parse
     * @param {ts.ScriptTarget} target TypeScript compiler script target
     * @param {Delinter} [delinter] (Optional) Delinter instance to use for delinting
     * @returns {uml.CodeModel} The parse results
     *
     * @memberOf TypeScriptUml
     */
    static parseFile(fileName, target, sourceText, delinter) {
        if (!delinter) {
            delinter = new delint_1.Delinter();
        }
        if (!sourceText) {
            sourceText = fs_1.readFileSync(fileName).toString();
        }
        const sourceFile = ts.createSourceFile(fileName, sourceText, target, /*setParentNodes */ true);
        delinter.parse(sourceFile);
        return delinter.umlCodeModel;
    }
    /**
     * Generate a uml class diagram from a uml CodeModel description.
     *
     * @static
     * @param {uml.CodeModel} codeModel Uml CodeModel description
     * @param {ITypeScriptUmlOptions} [options] Options
     * @returns {string} Class diagram formatted according to the specified options
     *
     * @memberOf TypeScriptUml
     */
    static generateClassDiagram(codeModel, options) {
        let _formatter = null;
        const defaultOptions = {
            formatter: "yuml",
            nodes: {
                exclude: null,
                include: null,
            },
            plantuml: {
                diagramTags: true,
            },
        };
        if (!options) {
            options = {};
        }
        if (!options.hasOwnProperty("formatter")) {
            options.formatter = defaultOptions.formatter;
        }
        if (!options.hasOwnProperty("nodes")) {
            options.nodes = defaultOptions.nodes;
        }
        else {
            if (!options.nodes.hasOwnProperty("include")) {
                options.nodes.include = defaultOptions.nodes.include;
            }
            if (!options.nodes.hasOwnProperty("exclude")) {
                options.nodes.exclude = defaultOptions.nodes.exclude;
            }
        }
        if (!options.hasOwnProperty("plantuml")) {
            options.plantuml = defaultOptions.plantuml;
        }
        else {
            if (!options.plantuml.hasOwnProperty("diagramTags")) {
                options.plantuml.diagramTags = defaultOptions.plantuml.diagramTags;
            }
        }
        let diagramCodeModel = codeModel;
        if (options.nodes.include) {
            // Include nodes
            diagramCodeModel = this._createIncludeCodeModel(codeModel, options.nodes.include, options.nodes.exclude);
        }
        else if (options.nodes.exclude) {
            // Exclude nodes
            diagramCodeModel = this._createExcludeCodeModel(codeModel, options.nodes.exclude);
        }
        switch (options.formatter) {
            case "yuml":
                _formatter = new formatter.YumlFormatter(options);
                break;
            case "plantuml":
                _formatter = new formatter.PlantUMLFormatter(options);
                break;
            default:
                throw new Error(`Unknown formatter ${options.formatter}`);
        }
        return _formatter.generateClassDiagram(diagramCodeModel);
    }
    /**
     * Fill options object with default values for unspecified properties.
     *
     * @private
     * @static
     * @param {IParseOptions} [options] Options object
     * @returns {IParseOptions} Options object with default values for unspecified properties
     *
     * @memberOf TypeScriptUml
     */
    static _setDefaultParseOptions(options) {
        const defaultOptions = {
            exclude: [],
            include: null,
            tsconfig: null,
        };
        if (!options) {
            options = defaultOptions;
        }
        Object.keys(defaultOptions).forEach((key) => {
            if (!options.hasOwnProperty(key)) {
                options[key] = defaultOptions[key];
            }
        });
        return options;
    }
    /**
     * Find and parse the tsconfig.json file for the project in the searchPath.
     *
     * @private
     * @static
     * @param {string} searchPath Base search path to look for the tsconfig
     * @param {string} options Path of the tsconfig.json file to parse (optional)
     * @returns {ts.ParsedCommandLine} The parse results
     *
     * @memberOf TypeScriptUml
     */
    static _readTsconfig(searchPath, options) {
        const configPath = options.tsconfig ? options.tsconfig : ts.findConfigFile(searchPath, fs_1.existsSync);
        const config = ts.readConfigFile(configPath, (path) => {
            return fs_1.readFileSync(path).toString();
        });
        if (config.error) {
            const formattedDiagnostic = ts.formatDiagnostics([config.error], this._defaultFormatDiagnosticsHost);
            throw new Error(`Failed to read tsconfig file ${configPath}: ${formattedDiagnostic}`);
        }
        if (options.include) {
            config.config.include = options.include;
        }
        if (options.exclude) {
            if (!config.config.exclude) {
                config.config.exclude = options.exclude;
            }
            else {
                config.config.exclude = config.config.exclude.concat(options.exclude);
            }
        }
        const parsed = ts.parseJsonConfigFileContent(config.config, ts.sys, searchPath);
        if (parsed.errors.length > 0) {
            const formattedDiagnostic = ts.formatDiagnostics(parsed.errors, this._defaultFormatDiagnosticsHost);
            throw new Error(`Failed to parse tsconfig file ${configPath}: ${formattedDiagnostic}`);
        }
        return parsed;
    }
    static _createIncludeCodeModel(codeModel, include, exclude) {
        const newCodeModel = new uml.CodeModel();
        if (exclude === null || exclude === undefined) {
            exclude = [];
        }
        codeModel.nodes.forEach((key, node) => {
            if (this._isIncluded(node, include) && !this._isExcluded(node, exclude)) {
                newCodeModel.nodes.setValue(key, codeModel.nodes.getValue(key));
            }
        });
        codeModel.associations.forEach((association) => {
            if (this._isIncluded(association, include) && !this._isExcluded(association, exclude)) {
                newCodeModel.associations.add(association);
            }
        });
        codeModel.generalizations.forEach((generalization) => {
            if (this._isIncluded(generalization, include) && !this._isExcluded(generalization, exclude)) {
                newCodeModel.generalizations.add(generalization);
            }
        });
        return newCodeModel;
    }
    static _createExcludeCodeModel(codeModel, exclude) {
        const newCodeModel = new uml.CodeModel();
        codeModel.nodes.forEach((key, node) => {
            if (!this._isExcluded(node, exclude)) {
                newCodeModel.nodes.setValue(key, codeModel.nodes.getValue(key));
            }
        });
        codeModel.associations.forEach((association) => {
            if (!this._isExcluded(association, exclude)) {
                newCodeModel.associations.add(association);
            }
        });
        codeModel.generalizations.forEach((generalization) => {
            if (!this._isExcluded(generalization, exclude)) {
                newCodeModel.generalizations.add(generalization);
            }
        });
        return newCodeModel;
    }
    static _isIncluded(object, include) {
        const node = object instanceof uml.Node ? object : null;
        const link = object instanceof uml.Link ? object : null;
        if (node) {
            return include.find((value) => {
                return node.identifier === value;
            }) !== undefined;
        }
        else if (link) {
            return include.find((value) => value === link.fromName) !== undefined &&
                include.find((value) => value === link.toName) !== undefined;
        }
    }
    static _isExcluded(object, exclude) {
        const node = object instanceof uml.Node ? object : null;
        const link = object instanceof uml.Link ? object : null;
        return exclude.find((value) => {
            if (node) {
                return node.identifier === value;
            }
            else if (link) {
                return link.fromName === value || link.toName === value;
            }
        }) !== undefined;
    }
}
/* istanbul ignore next: code only used by TypeScript API, which is mocked during tests */
TypeScriptUml._defaultFormatDiagnosticsHost = {
    getCanonicalFileName: (fileName) => {
        return ts.sys.useCaseSensitiveFileNames ? fileName : fileName.toLowerCase();
    },
    getCurrentDirectory: () => ts.sys.getCurrentDirectory(),
    getNewLine: () => ts.sys.newLine,
};
exports.TypeScriptUml = TypeScriptUml;
